local defaultPlayerShape = { l = -25, r = 25, b = -16, t = 96 }

local function GenericDeath(player)
	player.dead = true
	eapi.pointerMap[player.shape] = nil	
	eapi.Destroy(player.shape)
	player.shape = nil
	player.DisableInput()
end

local function FallDeath(player)
	GenericDeath(player)

	local shape = defaultPlayerShape
	player.shape = eapi.NewShape(player.body, nil, shape, "Fragment")
	eapi.pointerMap[player.shape] = player
	object.CompleteHalt(player)
	object.SimpleStep(player)
	
	local dieName = { "mslug/die.png", filter = true }
	local fallAnim = eapi.TextureToSpriteList(dieName, {55, 51})
	eapi.Destroy(player.tile)
	player.tile = eapi.NewTile(player.body, {-68,-28},
				   {137, 128}, fallAnim, 0)
	eapi.SetAttributes(player.tile, {flip = {player.direction, false}})
	eapi.Animate(player.tile, eapi.ANIM_CLAMP, 16)
	eapi.PlaySound(gameWorld, "mslug/fail.ogg", 0, 1.0)
end

local function PlayerStep(world, body)
	local player = eapi.pointerMap[body]
	local pos = eapi.GetPos(body)
	
	local worldData = eapi.GetData(world)
	local now = worldData.now
	local stepSec = worldData.stepSec
	
	local moveLeft = false
	local moveRight = false

	if player.ducked or (player.control.left and player.control.right) then
		moveLeft, moveRight = false, false
	else
		moveLeft = player.control.left
		moveRight = player.control.right
	end
	
	-- Left/right movement.
	if moveRight then
		eapi.SetAttributes(player.tile, {flip = {true, false}})
		player.direction = false
		if player.contact.ground then
			-- Walking on ground.
			player.vel.x = player.vel.x + 10
			if player.vel.x > player.maxWalkSpeed then
				player.vel.x = player.maxWalkSpeed
			end
		else
			-- Flying through air.
			player.vel.x = player.vel.x + 5
			if player.vel.x > player.maxWalkSpeed then
				player.vel.x = player.maxWalkSpeed
			end
		end
	elseif player.vel.x > 0 and not moveLeft then
		if player.contact.ground then
			-- Ground friction.
			player.vel.x = player.vel.x - 5
			if player.vel.x < 0 then
				player.vel.x = 0
			end
		else
			-- Air resistance.
			player.vel.x = 0.99 * player.vel.x
			if player.vel.x < 0 then
				player.vel.x = 0
			end
		end
	end
	if moveLeft then
		eapi.SetAttributes(player.tile, {flip = {false, false}})
		player.direction = true
		if player.contact.ground then
			-- Walking on ground.
			player.vel.x = player.vel.x - 10
			if player.vel.x < -player.maxWalkSpeed then
				player.vel.x = -player.maxWalkSpeed
			end
		else
			-- Flying through air.
			player.vel.x = player.vel.x - 5
			if player.vel.x < -player.maxWalkSpeed then
				player.vel.x = -player.maxWalkSpeed
			end
		end
	elseif player.vel.x < 0 and not moveRight then
		if player.contact.ground then
			-- Ground friction.
			player.vel.x = player.vel.x + 5
			if player.vel.x > 0 then
				player.vel.x = 0
			end
		else
			-- Air resistance.
			player.vel.x = 0.99 * player.vel.x
			if player.vel.x > 0 then
				player.vel.x = 0
			end
		end
	end
	
	-- Apply downward velocity in case we're moving down a slope.
	local slope = player.contact.leftSlope
	if slope and player.vel.x < 0 then
		-- We're moving left on a left-slope.
		local num = math.floor(player.vel.x * stepSec) * slope.h
		local den = (slope.w * stepSec)
		player.vel.y = num / den
	end
	slope = player.contact.rightSlope
	if slope and player.vel.x > 0 then
		-- We're moving right on a right-slope.
		local num = math.floor(player.vel.x * stepSec) * slope.h
		local den = (slope.w * stepSec)
		player.vel.y = -num / den
	end
	
	-- Jumping.
	if player.control.jump then
		if player.contact.onewayGround and player.control.down then
			player.downJump = true
			player.control.jump = false
		elseif player.contact.ground 
			and not player.ducked
			and not player.jumpInProgress then
			player.vel.y = 600
			player.jumpInProgress = true
			player.control.jump = false
		end
	end
	
	-- Apply gravity to velocity.
	player.vel = vector.Add(player.vel, vector.Scale(player.gravity, stepSec))
	
	-- Update player position.
	local newPos = vector.Add(pos, vector.Scale(player.vel, stepSec))
	eapi.SetPos(player.body, newPos)
	
	-- Save and reset contact state.
	player.prevContact = player.contact
	player.contact = {}
end

local function NewPlayerTile(player, anim)
	if anim == player.currentAnim or player.stopped then return end
	if player.tile then eapi.Destroy(player.tile) end
	player.tile = eapi.NewTile(player.body, {-45,-28}, {90, 128}, anim, 0)
	player.currentAnim = anim
end

local function PlayerAfterStep(world, body)
	local player = eapi.pointerMap[body]
	local bodyData = eapi.GetData(body)
	local tileAttr = eapi.GetAttributes(player.tile)
	
	-- If we were in contact with a platform before, but not anymore, unlink
	-- player.
	if player.prevContact.platform and (not player.contact.platform) then
		eapi.Unlink(player.body)
	end
	
	-- Note: we compute the change in position ourselves because the
	-- .deltaPos value provided by engine is the actual change in (rounded)
	-- position. Even if position did not change due to rouning in previous
	-- step, player might still be moving and must be animated accordingly.
	local deltaPos = vector.Sub(bodyData.pos, bodyData.prevPos)
	
	-- Assume we're not moving if touching left or right wall.
	if (player.contact.left or player.contact.right) then
		deltaPos.x = 0
	end
	
	if player.contact.ground then
		-- Walking animation.
		if deltaPos.x ~= 0 then
			local dir = -1
			if ((player.vel.x > 0 and not player.direction) or
		           ((player.vel.x < 0 and player.direction)))  then
				dir = 1
			end
			NewPlayerTile(player, player.walkAnim)
			eapi.SetAnimPos(player.tile, player.animPos)
			local tmp = player.animSpeed * math.abs(deltaPos.x)
			player.animPos = player.animPos + dir * tmp
		else
			-- Pick the frame where player looks as if standing still.			
			if not(player.ducked) then
				NewPlayerTile(player, player.standAnim)
				local dir = not(player.direction or false)
				local data = {flip={dir, false}}
				eapi.SetAttributes(player.tile, data)
			end
		end
	end
	
	-- .downJump is used to signal one-way-ground collision handler that
	-- player must be allowed to fall through. Reset this value since
	-- collision handlers have been executed already.
	player.downJump = false
	
	-- Make sure player has to hit jump-key every time.
	if not player.contact.ground or (not player.jumpInProgress) then
		player.control.jump = false
	end

	if player.activator then
		if player.activator.active then
			player.activator.active = false
		else
			action.RemoveActivatorInfo(player.activator)
			player.activator = nil
		end
	end	
end

local function CreatePlayerShape(player, shouldDestroy, shape)
	shape = shape or defaultPlayerShape
	if player.shape then
		eapi.pointerMap[player.shape] = nil
		if shouldDestroy then
			eapi.Destroy(player.shape)
		end
	end
	player.shape = eapi.NewShape(player.body, nil, shape, "Player")
	eapi.pointerMap[player.shape] = player
end

--[[
	Make our destroyer player character appear engine-side:
		* load player Body, Shape, and Tile objects,
		* bind movement keys,
		* set step and after-step functions.
]]--

local ActivateLock = false

local function Load(world, player)
	player.activator = nil
	ActivateLock = false
	player.shootInProgress = false
	player.jumpInProgress = false	-- reset jumping state.
	player.maxWalkSpeed = 400
	player.id = "player"
	player.handTile = nil
	player.gunTile = nil	
	player.ducked = nil
	player.aimUp = nil
	player.dead = false
	player.invincible = false
	player.direction = true

	-- Clear out stale body references, and create a new body.
	if player.body then
		eapi.pointerMap[player.body] = nil
	end
	player.body = eapi.NewBody(world, {0,0})
	eapi.pointerMap[player.body] = player

	-- Clear out stale shape references, and create a new shape.
	CreatePlayerShape(player)
	
	-- Velocity and gravity.
	player.vel = {x=0,y=0}
	player.gravity = {x=0,y=-1500}
	
	-- Animations.
	local function mslugSprite(name)
		local size = {33, 47}
		name = "mslug/" .. name
		return eapi.TextureToSpriteList({ name, filter = true }, size)
	end
	player.walkAnim = mslugSprite("run.png")
	player.standAnim = mslugSprite("stand.png")

	local function MoveLeft(key, keyDown)
		player.control.left = keyDown
	end
	local function MoveRight(key, keyDown)
		player.control.right = keyDown
	end
	local function Jump(key, keyDown)
		player.control.jump = keyDown
	end

	player.EnableInput = function()
		player.inputDisabled = false
		eapi.SetStepFunc(player.body, PlayerStep, PlayerAfterStep)
	end

	player.DisableInput = function()
		player.inputDisabled = true
		eapi.SetStepFunc(player.body, nil, nil)
	end

	-- Bind keys.	
	player.StartInput = function ()
		eapi.BindKey(eapi.KEY_LEFT, MoveLeft)
		eapi.BindKey(eapi.KEY_RIGHT, MoveRight)
		eapi.BindKey(eapi.KEY_z, util.MakeMessageFinish(Jump))
	end

	player.StopInput = function ()
		player.control = {}
		local skipFn = util.MakeMessageFinish()
		eapi.BindKey(eapi.KEY_LEFT)
		eapi.BindKey(eapi.KEY_RIGHT)
		eapi.BindKey(eapi.KEY_z, skipFn)
	end

	-- Create main player tile and set its sprite list.
	NewPlayerTile(player, player.walkAnim)
		
	player.StartInput()
		
	-- Set step functions.
	player.EnableInput()

	object.VerticalHalt(player)

	-- Register collision handlers.
	object.RegisterHandlers("Player")
end

local function Create()
	local player = {}
	player.type = "Destroyer"
	player.contact = {}
	player.prevContact = {}
	player.control = {}

	-- Constants.
	player.animSpeed = 0.005
	player.animPos = 0.0

	return player
end

local function Place(player, position, flipH)
	eapi.SetPos(player.body, position)
	NewPlayerTile(player, player.standAnim)
	player.direction = flipH
	if flipH then
		eapi.SetAttributes(player.tile, {flip={true, false}})
	end
end

local function Restore(player)
	eapi.StopAnimation(player.tile)
	eapi.SetPos(player.tile, {-32,-28})
	eapi.SetSpriteList(player.tile, player.walkAnim)
	player.EnableInput()
end
	
function Activate(DoTheActualActivation, DoPostActivation)
	local function SecondAfterAnim()
		Restore(mainPC)
		ActivateLock = false
		if DoPostActivation then
			DoPostActivation()
		end
	end
	local function Continue()
		eapi.Animate(mainPC.tile, eapi.ANIM_CLAMP, -24)
		eapi.AddTimer(mainPC.body, 0.333, SecondAfterAnim)
	end
	local function AfterAnim()
		local shouldWait = false
		if DoTheActualActivation then
			shouldWait = DoTheActualActivation(Continue)
		end
		if not(shouldWait) then 
			Continue()
		end
	end
	if not(ActivateLock) then
		ActivateLock = true
		destroyer.Turn(mainPC)
		eapi.AddTimer(mainPC.body, 0.333, AfterAnim)
	end
end

destroyer = {
	FallDeath = FallDeath,
	Activate = Activate,
	Restore = Restore,
	Create = Create,
	Load = Load,
	Place = Place,
}
return destroyer
