--[[
	Various utility routines.
--]]

Infinity = 1e+308
smallTileSize = {32, 32}
defaultTileSize = {64, 64}
defaultFontset = nil

showFPS = false

--[[
	String split function from http://lua-users.org/wiki/SplitJoin
]]--
local function Split(str, delim, maxNb)
	-- Eliminate bad cases...
	if string.find(str, delim) == nil then
		return { str }
	end
	if maxNb == nil or maxNb < 1 then
		maxNb = 0    -- No limit
	end
	local result = {}
	local pat = "(.-)" .. delim .. "()"
	local nb = 0
	local lastPos
	for part, pos in string.gfind(str, pat) do
		nb = nb + 1
		result[nb] = part
		lastPos = pos
		if nb == maxNb then break end
	end
	-- Handle the last field
	if nb ~= maxNb then
		result[nb + 1] = string.sub(str, lastPos)
	end
	return result
end

local function ValueInTable(value, table)
	for k, v in pairs(table) do
		if v == value then
			return true
		end
	end
	return false
end

local function Indent(level)
	for i = 2,level do
		io.write("\t")
	end
end

local function PrintTable(tbl, level)
	if not level then
		level = 1
	end

	-- opening bracket
	Indent(level)
	io.write("{\n")
	
	for k, v in pairs(tbl) do
		if type(k) == "table" then
			PrintTable(k, level + 1)
			Indent(level + 1)
			io.write(": ")
		else
			Indent(level + 1)
			io.write(tostring(k), ": ")
		end
		
		if type(v) == "table" then
			PrintTable(v, level + 1)
		else
			io.write(tostring(v), "\n")
		end
	end
	Indent(level)
	io.write("}\n")
end

--[[ Number rounding function from http://lua-users.org/wiki/SimpleRound ]]--
local function Round(num, idp)
	local mult = 10^(idp or 0)
	return math.floor(num * mult + 0.5) / mult
end

--[[
	Accepts an array of tile specifications and executes eapi.NewTile() for
	each one of them.

	tiles		Tile data -- an array with each element in the
			form {{X, Y}, sprite}
	tileSize	The tile grid coordinates {X, Y} above are multiplied by
			tile size to get the actual world coordinates. If not
			specified, assumed to be defaultTileSize.
]]--
local function TileBatch(tiles, tileSize)
	tileSize = tileSize or defaultTileSize
	for i, v in ipairs(tiles) do
		local pos = v[1]
		pos = { pos[1] * tileSize[1], pos[2] * tileSize[2] }
		local canvas = {l=0, r=tileSize[1], b=0, t=tileSize[2]}
		eapi.NewTile(nil, pos, canvas, v[2])
	end
end

--[[
	Given a tile map (represented in ASCII) and a tile set
	(ASCII character to sprite mapping) execute eapi.NewTile() for each
	character to make the tiles appear in the game world.

	body		Body object the tiles will be attached to.
	tileMap		Array of strings.
	tileSet		String character to sprite list mapping.
	startPos	Position of the bottom left tile. nil means {0,0}.
	tileSize	Size of the created tiles. Character row and column
			position is multiplied by this and added to startPos to
			get the final world coordinate. nil means
			defaultTileSize.
	depth		Depth, nil = 0.
]]--
local function CreateTiles(body, tileMap, tileset, startPos, tileSize, depth)
	tileSize = tileSize or defaultTileSize
	startPos = startPos or {x=0,y=0}
	depth = depth or 0
	for r, charRow in ipairs(tileMap) do
		for c = 1, #charRow do
			local char = string.sub(charRow, c, c)
			local sprite = tileset[char]
			assert(char == " " or sprite, "Character (" .. char ..
			   ") not found in tileMap.")
			if sprite then
				local pos = {
				    x=startPos.x + (c-1)*tileSize[1],
				    y=startPos.y + (#tileMap-r)*tileSize[2]
				}
				eapi.NewTile(body, pos, tileSize, sprite, depth)
			end
		end
	end
end

function RandomElement(list)
	return list[util.Random(1,#list)]
end

function xor (x, y)
	return ((x and not(y)) or (not(x) and y))
end

local function PutTile(body, tileID, tileSet, position, depth)
	return eapi.NewTile(body, position, nil, tileSet[tileID], depth)
end

local function PutTileWithAttribute(body, tileID, tileSet, position, depth, attr)
	local tile = eapi.NewTile(body, position, nil, tileSet[tileID], depth)
	if attr then
		eapi.SetAttributes(tile, attr)
	end
end

local function PutAnimTile(body, spritelist, position, depth, animType, FPS, animStartTime)
	local tile = eapi.NewTile(body, position, nil, spritelist, depth)
	eapi.Animate(tile, animType, FPS, animStartTime)
	return tile
end

local function PutSmallTile(body, tileID, tileSet, position, depth)	 
	 eapi.NewTile(body,
                      position,
	              smallTileSize,
	              tileSet[tileID],
                      depth)
end 

--[[
	Chop texture into sprites to create a tileset.

	texture		Image file name or texture object.
	spriteMap	Character rows that "name" each sprite (see below).
	spriteSize	Size of each sprite in pixels.

	Each character in spriteMap corresponds to a rectangular area of the
	texture (a sprite). Returned result is a tileset: a mapping from each
	character that appears in spriteMap to a SpriteList object.

	If a character appears multiple times in the spriteMap, its SpriteList
	has multiple sprites (frames), which is a useful way to group sprites
	for animation.

	So this:
	spriteMap = {
		"abccd ",
		"   xxy"
	}
	
	produces this:
	tileset = {
		a = eapi.NewSpriteList(texture, {{0,0},{32,32}}),
		b = eapi.NewSpriteList(texture, {{32,0},{32,32}}),
		c = eapi.NewSpriteList(texture, {{64,0},{32,32}}, {{96,0},{32,32}}),
		...
	}
]]--
local function TextureToTileset(texture, spriteMap, spriteSize, offset)
	local tileset = {}
	offset = offset or {x=0, y=0}
	
	-- Figure out sprite coordinates and store them in the tileset table.
	for r, charRow in ipairs(spriteMap) do
		for c = 1, #charRow do
			local char = string.sub(charRow, c, c)
			if char ~= " " then
				if tileset[char] == nil then
					tileset[char] = {}
				end
				local texfrag = {
				  {(c-1)*spriteSize[1] + offset.x,
				   (r-1)*spriteSize[2] + offset.y},
				  {spriteSize[1], spriteSize[2]}
				}
				table.insert(tileset[char], texfrag)
			end
		end
	end

	-- Turn sprite coordinates into actual sprite list objects.
	for char, texfrag in pairs(tileset) do
		tileset[char] = eapi.NewSpriteList(texture, unpack(texfrag))
	end
	
	return tileset
end

--[[
	Load font texture, chop it into sprites, and return the SpriteList.

	filename	Image filename.
	spriteSize	Size of each character in pixels.
	
	In the font image, space (32 ascii) character is expected to be the
	first, and tilde (126 ascii) the last. The ordering is left to right.
	If split over multiple lines, the lines are ordered top to bottom.
]]--
local function LoadFont(filename, spriteSize)
	local fontset = {
		spriteList=eapi.TextureToSpriteList(filename, spriteSize),
		spriteSize=spriteSize
	}
	return fontset
end

local fontMap = { }
fontMap[string.char(0xc4,0x80)] = string.char(128) --Ā
fontMap[string.char(0xc4,0x8c)] = string.char(129) --Č
fontMap[string.char(0xc4,0x92)] = string.char(130) --Ē
fontMap[string.char(0xc4,0xa2)] = string.char(131) --Ģ
fontMap[string.char(0xc4,0xaa)] = string.char(132) --Ī
fontMap[string.char(0xc4,0xb6)] = string.char(133) --Ķ
fontMap[string.char(0xc4,0xbb)] = string.char(134) --Ļ
fontMap[string.char(0xc5,0x85)] = string.char(135) --Ņ
fontMap[string.char(0xc5,0xa0)] = string.char(136) --Š
fontMap[string.char(0xc5,0xaa)] = string.char(137) --Ū
fontMap[string.char(0xc5,0xbd)] = string.char(138) --Ž

fontMap[string.char(0xc4,0x81)] = string.char(144) --ā
fontMap[string.char(0xc4,0x8d)] = string.char(145) --č
fontMap[string.char(0xc4,0x93)] = string.char(146) --ē
fontMap[string.char(0xc4,0xa3)] = string.char(147) --ģ
fontMap[string.char(0xc4,0xab)] = string.char(148) --ī
fontMap[string.char(0xc4,0xb7)] = string.char(149) --ķ
fontMap[string.char(0xc4,0xbc)] = string.char(150) --ļ
fontMap[string.char(0xc5,0x86)] = string.char(151) --ņ
fontMap[string.char(0xc5,0xa1)] = string.char(152) --š
fontMap[string.char(0xc5,0xab)] = string.char(153) --ū
fontMap[string.char(0xc5,0xbe)] = string.char(154) --ž


local function ConvertFromUnicode(text)
	local i = 1
	local newText = ""
	repeat
		local two = string.sub(text, i, i + 1)
		local x = fontMap[two]
		if x then
			newText = newText .. x
			i = i + 2
		else
			newText = newText .. string.sub(text, i, i)
			i = i + 1
		end
	until #two < 2
	return newText
end

--[[
	Create text tiles that are attached to camera viewport.

	text		Text string.
	camera		Camera pointer.
	screenPos	Screen is mapped like so:
			{-1, 1}   {0, 1}   {1, 1}

			{-1, 0}   {0, 0}   {1, 0}

			{-1,-1}   {0,-1}   {1,-1}

			If nil, {0,0} (screen center) is used. Other values than
			-1, 0, 1 are possible and produce in-between positions.
	depth		Depth determines drawing order. Defaults to 100.
	fontset		Font sprite list as returned by LoadFont(). If nil,
			defaultFontset is used.
--]]

local black = { r = 0, g = 0, b = 0 }

local function CreateViewportLines(lines, camera, pos, depth, fontset)
	fontset = fontset or defaultFontset
	assert(fontset, "Font tileset not loaded.")
	depth = depth or 100

	local function PutChar(p, color, z)
		local tile = eapi.NewTile(camera, p, nil, fontset.spriteList, z)
		if color then eapi.SetAttributes(tile, {color = color}) end
		return tile
	end

	local ink = nil
	local textTiles = {}
	local camSize = eapi.GetSize(camera)
	pos.x = pos.x - camSize.x / 2 
	pos.y = camSize.y / 2 - fontset.spriteSize[2] - pos.y
	for r, line in ipairs(lines) do
		if type(line) == "table" then
			ink = line.ink
			line = line.str
		else
			ink = nil
		end
		local leftSide = pos.x
		for c = 1, #line do
			local char = string.sub(line, c, c)
			local rpos = vector.Round(pos)
			local tile1 = PutChar(rpos, ink, depth)
			local shadowPos = vector.Add(rpos, {x=1,y=-1})
			local tile2 = PutChar(shadowPos, black, depth - 1)
			
			-- Choose the right sprite (frame) from list for char.
			char = string.byte(char, 1)
			if char < 32 or char > 154 then
				print("Character", char, "not in fontset.")
				char = string.byte("?", 1)
			end
			eapi.SetFrame(tile1, char - 32);
			eapi.SetFrame(tile2, char - 32);
			
			pos.x = pos.x + fontset.spriteSize[1]
			table.insert(textTiles, tile1)
			table.insert(textTiles, tile2)
		end
		pos.x = leftSide
		pos.y = pos.y - fontset.spriteSize[2]
	end

	return textTiles
end

local function CreateViewportText(text, camera, screenPos, depth, fontset)
	return CreateViewportLines(Split(ConvertFromUnicode(text), "\n"),
				   camera, screenPos, depth, fontset)
end

local function MaxWidth(lines)
	local maxWidth = 0
	for i, line in ipairs(lines) do
		if type(line) == "table" then 
			lines[i].str = ConvertFromUnicode(line.str)
			maxWidth = math.max(#lines[i].str, maxWidth)
		else
			lines[i] = ConvertFromUnicode(line)
			maxWidth = math.max(#lines[i], maxWidth)
		end
	end
	return maxWidth
end		
		
local dialogColor = { r = 0.35, g = 0.35, b = 0.35, a = 0.65 }

local function GameMsgLoc(totalSize, boxSize, axis)
	return  (totalSize - boxSize) / ((axis == "x" and 2) or 4)
end

local function DialogBox(lines, camera, calcPos)
	calcPos = calcPos or GameMsgLoc
	local camSize = eapi.GetSize(camera.ptr)
	local boxW = defaultFontset.spriteSize[1] * MaxWidth(lines) + 16
	local boxH = defaultFontset.spriteSize[2] * #lines + 16
	local x = calcPos(camSize.x, boxW, "x")
	local y = calcPos(camSize.y, boxH, "y")
	local attr = { size = { boxW,boxH }, color = dialogColor }
	local occTile = Occlusion.put('h', x - camSize.x / 2 - 8,
				      camSize.y / 2 - boxH - y + 8,
				      98, attr, camera.ptr)
	local tiles = CreateViewportLines(lines, camera.ptr, { x = x, y = y })
	table.insert(tiles, occTile)
	return tiles
end

local emptyDummy = { }
local dialogBoxShowing = false
local dialogBoxQueue = { tail = emptyDummy, head = emptyDummy }

local function GameMessage(lines, camera, Trigger)
	if dialogBoxShowing then
		dialogBoxQueue.tail.next = { }
		dialogBoxQueue.tail.lines = lines
		dialogBoxQueue.tail.camera = camera
		dialogBoxQueue.tail.trigger = Trigger
		dialogBoxQueue.tail = dialogBoxQueue.tail.next
	else
		dialogBoxShowing = { }
		dialogBoxShowing.trigger = Trigger
		dialogBoxShowing.tiles = DialogBox(lines, camera)
	end
end

local function MessageDone()
	if dialogBoxShowing then
		local trigger = dialogBoxShowing.trigger
		eapi.Destroy(unpack(dialogBoxShowing.tiles))
		dialogBoxShowing = false
		if trigger then trigger() end
		if dialogBoxQueue.head.lines then
			local head = dialogBoxQueue.head
			dialogBoxQueue.head = dialogBoxQueue.head.next
			GameMessage(head.lines, head.camera, head.trigger)
		end
	end
end

local function MakeMessageFinish(fn)
	return function(key, keyDown)
		if keyDown then
			MessageDone()
		end
		if fn then 
			fn(key, keyDown)
		end
	end
end

local function GameInfoLocation(totalSize, boxSize, axis)
	return totalSize - ((axis == "x" and boxSize) or 32)
end

local function GameInfo(line, camera)
	return DialogBox({ line }, camera, GameInfoLocation)	
end

--[[
	Create FPS text tiles and return an array of their pointers.

	camera		Camera pointer as returned by eapi.NewCamera().
]]--
local function CreateFPSText(camera)
	local FPS, FPSText
	
	FPS = eapi.GetFPS()
	FPSText = "FPS: " .. FPS
	FPSText = FPSText .. string.rep(" ", 9-#FPSText)

	return CreateViewportText(FPSText, camera, {x=16, y=16})
end

local CameraTracking = { }

local function CreateCamera(world, obj, boundary, SomeFunction)
	local camera = {}
	local pos = nil
	local track = nil
	local offset = 0
	local first = true
	local stops = { }
	local stopRange = 275

	CameraTracking.call = function (value)
		track = value
	end

	CameraTracking.stop = function (id, value)
		stops[id] = value
	end
	
	if doNotInitCamera then return end
	if type(obj) == "table" then
		if obj.id == "player" then
			pos = eapi.GetPos(obj.body)
			track = obj
		else
			pos = obj
		end
	end
	
	camera.ptr = eapi.NewCamera(world, pos, nil, nil)
	camera.size = eapi.GetSize(camera.ptr)
	if boundary then
		eapi.SetBoundary(camera.ptr, boundary)
	end

	local function CameraAfterStep()
		-- In after-step function, we get the updated body position
		-- after collision handlers have run.
		if track and first then
			eapi.SetPos(camera.ptr, eapi.GetPos(track.body))
			first = false
		elseif track then
			local s = eapi.GetSize(camera.ptr)
			local p = eapi.GetPos(track.body)
			if track.w then p.x = p.x + 0.5 * track.w end
			local velx = track.vel.x
			if (velx < 0 and not(track.direction))
			or (velx > 0 and track.direction) then
				velx = -velx
			end
			local d = velx - offset
			local new_offset = offset + d * 0.005
			if (new_offset > offset and velx > 0)
			or (new_offset < offset and velx < 0) then
				offset = new_offset
			end
			p = vector.Add(p, {x = 0.9 * offset, y = 0.2 * s.y})
			local c = eapi.GetPos(camera.ptr)
			p = vector.Scale(vector.Sub(p, c), 0.03)

			-- this is camera jerking fix reported by Einars
			-- may this is not proper way to fix it, but it works
			if vector.Length(p) < 0.2 then
				p = { x = 0, y = 0 }
			end

			local p = vector.Add(c, p)
			-- stop tracking camera at certain points in scene
			for i in pairs(stops) do
				local x1 = vector.Distance(c, stops[i])
				local x2 = vector.Distance(p, stops[i])
				if x1 < stopRange and x2 <= x1 then return end
			end
			eapi.SetPos(camera.ptr, p)
		end
		
		if camera.FPSTextTiles then
			-- Destroy previous FPS text.
			eapi.Destroy(unpack(camera.FPSTextTiles))
			camera.FPSTextTiles = nil
		end
		if showFPS then
			camera.FPSTextTiles = CreateFPSText(camera.ptr)
		end

		if SomeFunction then
			SomeFunction()
		end
	end
	
	eapi.SetStepFunc(camera.ptr, nil, CameraAfterStep)

	return camera
end

local function BindDebugKeys()
	local function Reset(key, keyDown)
		if keyDown then
			eapi.Clear()
			dofile("script/first.lua")
		end
	end
	eapi.BindKey(eapi.KEY_F12, Reset)
end

local function CreateSimplePlatform(pos, velocity,
				    TilesAndShape,
				    controllable)
	local platform = {}
	platform.body = eapi.NewBody(gameWorld, pos)
	platform.vel = (controllable and {x=0, y=0}) or velocity
	local bounds = TilesAndShape(platform)

	local function PlatformStep(world)
		local worldData = eapi.GetData(world)
		local stepSec = worldData.stepSec
		local pos = eapi.GetPos(platform.body)

		local deltaPos = vector.Scale(platform.vel, stepSec)
		pos = vector.Add(pos, deltaPos)

		if (pos.x >= bounds.r or pos.y >= bounds.t)
		    or (pos.x <= bounds.l or pos.y <= bounds.b) then
			if controllable then 
				platform.vel = { x = 0, y = 0 }
				eapi.SetStepFunc(platform.body, nil, nil)
				if type(controllable) == "function" then
					controllable()
				end
			else
				platform.vel = vector.Reverse(platform.vel)
			end
		end

		-- Set new platform position, and get the actual moved (rounded)
		-- unit distance from engine.
		eapi.SetPos(platform.body, pos)
		deltaPos = eapi.GetDeltaPos(platform.body)
		
		-- Apply position difference to each child.
		local children = eapi.GetChildren(platform.body)
		for index, child in ipairs(children) do
			local childPos = eapi.GetPos(child)
			eapi.SetPos(child, vector.Add(childPos, deltaPos))
		end
	end

	local function PlatformUp()
		platform.vel = velocity
		eapi.SetStepFunc(platform.body, PlatformStep, nil)
	end
	local function PlatformDown()
		platform.vel = vector.Reverse(velocity)
		eapi.SetStepFunc(platform.body, PlatformStep, nil)
	end

	eapi.pointerMap[platform.shape] = platform;
	eapi.SetStepFunc(platform.body, PlatformStep, nil)

	return { up = PlatformUp, down = PlatformDown }
end

local function ToBeOrNotToBe()
	if (util.Random() > 0.5) then
		return true
	else
		return false
	end		
end

tileMap8x8 = { "abcdefgh",
	       "ijklmnop",
	       "qrstuvwx",
	       "yzABCDEF",
	       "GHIJKLMN",
	       "OPQRSTUW",
	       "XYZ12345",
	       "67890!@#" }

local function ParticleInterval(avgInterval)
	-- maybe gausian distribution would be better, 
	-- but it involves lots of maths therefore for now
	-- just return random value in interval [ 0; 2*avgInterval ]
	return 2.0 * util.Random() * avgInterval
end

local function ParticleEmitter(origin, life, avgInterval, configure)
	local emitter = { }
	local function NewParticle()
		local particle = eapi.NewBody(gameWorld, origin())
		local function ParticleRipper()
			eapi.Destroy(particle)
		end
		configure(particle)
		eapi.AddTimer(particle, life, ParticleRipper)
		if emitter.active then
			local interval = ParticleInterval(avgInterval) 
			eapi.AddTimer(particle, interval, NewParticle)
		end
	end

	emitter.Kick = function()
		local oldState = emitter.active
		emitter.active = true
		if not(oldState) then
			NewParticle()
		end
	end

	emitter.Stop = function()
		emitter.active = false
	end

	return emitter
end

-- A function that is supposed to be bound to menu key (escape) and will display
-- the menu once executed.
local function ShowMenu(key, keyDown, gameOver)
	if not keyDown then
		return
	end
	
	-- Pause game world and save game keys.
	eapi.Pause(gameWorld)
	local keyBind = eapi.GetKeyBindings()
	
	if not gameOver then
		-- Bind a function to escape key that destroys the menu.
		local function HideMenu(key, keyDown)
			if key and (not keyDown) then
				return
			end
			
			-- Destroy menu and restore game key bindings.
			menu.Kill()
			eapi.SetKeyBindings(keyBind)
		end
		
		-- Display menu.
		menu.Show(HideMenu)
	else
		-- Game over, unbind escape key and display menu without resume
		-- option.
		menu.Show(nil)
	end
end

-- Game-over sequence.
local function GameOver()
	local keyBind = eapi.GetKeyBindings()
	eapi.SetKeyBindings({})		-- Unbind keys, disable all input.

	local world = eapi.NewWorld("Game Over", 5, 11)
	local camera = CreateCamera(world, nil, nil, nil)
	local dimLevel = 0
	local function DimLights()
		dimLevel = dimLevel + 0.01
		if dimLevel >= 0.5 then
			-- Destroy this "darkening" world, restore key
			-- bindings and show game-over menu.
			eapi.Destroy(world)
			eapi.SetKeyBindings(keyBind)
			ShowMenu(nil, true, true)
			return
		end
		eapi.SetBackgroundColor(world, {r=0,g=0,b=0,a=dimLevel})
		eapi.AddTimer(world, 0.02, DimLights)
	end
	eapi.AddTimer(world, 0.02, DimLights)
end

local Message = nil

-- Wipe engine-side and client-side state and execute a script file.
local function GoTo(roomName, TransitionFunc, disableESCAPE)
	eapi.Clear()			-- Clear state.
	
	emptyDummy = { }
	dialogBoxQueue = { tail = emptyDummy, head = emptyDummy }
	dialogBoxShowing = false
	
	-- Create game world with 5 millisecond step.
	gameWorld = eapi.NewWorld("Game", 5, 16)

	defaultFontset = LoadFont("image/default-font.png", {8,16})
	BindDebugKeys()

	-- Load engine font for displaying FPS, console output, etc.
	eapi.LoadEngineFont(defaultFontset.spriteList)

	object.RegisterHandlers("Fragment")
	dofile("script/" .. roomName .. ".lua")
	if TransitionFunc then
		TransitionFunc()
	end
	Message = nil
end

local function GoToAndPlace(room, player, pos, flip)
	local function PlacePC()
		destroyer.Place(player, pos, flip)
	end
	GoTo(room, PlacePC)
end

local function FileExists(n)
	local f = io.open(n)
	if f == nil then
		return false
	else
		io.close(f)
		return true
	end
end

local function randomerror(a, b)
	a = a or "nil"
	b = b or "nil"
	print("Error in unil.random(" .. a .. ", " .. b .. ")")
	eapi.Quit()
end

local function Random(a, b)
	local rnd = eapi.Random() / 2147483648
	if a then
		if not(b) then
			b = a
			a = 1
		end
		if (a > b) then 
			randomerror(a, b)
		end
		return math.floor(rnd * (b - a + 1)) + a
	elseif b then
		randomerror(a, b)
	else
		return rnd
	end
end

util = {
	BindDebugKeys = BindDebugKeys,
	CreateCamera = CreateCamera,
	CreateFPSText = CreateFPSText,
	CreateTiles = CreateTiles,
	CreateViewportText = CreateViewportText,
	FileExists = FileExists,
	GoTo = GoTo,
	GoToAndPlace = GoToAndPlace,
	GameOver = GameOver,
	ShowMenu = ShowMenu,
	PrintTable = PrintTable,
	PutAnimTile = PutAnimTile,
	PutSmallTile = PutSmallTile,
	PutTile = PutTile,
	PutTileWithAttribute = PutTileWithAttribute,
	Round = Round,
	Split = Split,
	TextureToTileset = TextureToTileset,
	TileBatch = TileBatch,
	ValueInTable = ValueInTable,
	ToBeOrNotToBe = ToBeOrNotToBe,
	ParticleEmitter = ParticleEmitter,
	CreateSimplePlatform = CreateSimplePlatform,
	GameMessage = GameMessage,
	GameInfo = GameInfo,
	Message = Message,
	MessageDone = MessageDone,
	MakeMessageFinish = MakeMessageFinish,
	Random = Random,
	map8x8 = tileMap8x8,
	CameraTracking = CameraTracking,
	DialogBox = DialogBox,
	dialogColor = dialogColor,
}
return util
